@@ -24,6 +24,16 @@ import pandas as pd
 from ..algorithm.leiden import leiden as le
 from ..algorithm._louvain import louvain as lo
 from typing_extensions import Literal
+from ..log_manager import logger
+
+
+def logit(func):
+    def wrapped(*args, **kwargs):
+        logger.info('start to run {}...'.format(func.__name__))
+        res = func(*args, **kwargs)
+        logger.info('{} end.'.format(func.__name__))
+        return res
+    return wrapped
 
 
 class StPipeline(object):
@@ -36,6 +46,7 @@ class StPipeline(object):
         self.data = data
         self.result = dict()
         self._raw = None
+        self.key_record = {'hvg': [], 'pca': [], 'neighbors': [], 'umap': [], 'cluster': [], 'marker_genes': []}
 
     @property
     def raw(self):
@@ -67,6 +78,21 @@ class StPipeline(object):
     def raw_checkpoint(self):
         self.raw = self.data
 
+    def reset_key_record(self, key, res_key):
+        """
+        reset key and coordinated res_key in key_record.
+        :param key:
+        :param res_key:
+        :return:
+        """
+        if key in self.key_record.keys():
+            if res_key in self.key_record[key]:
+                self.key_record[key].remove(res_key)
+            self.key_record[key].append(res_key)
+        else:
+            self.key_record[key] = [res_key]
+
+    @logit
     def cal_qc(self):
         """
         calculate three qc index including the number of genes expressed in the count matrix, the total counts per cell
@@ -76,6 +102,7 @@ class StPipeline(object):
         """
         cal_qc(self.data)
 
+    @logit
     def filter_cells(self, min_gene=None, max_gene=None, min_n_genes_by_counts=None, max_n_genes_by_counts=None,
                      pct_counts_mt=None, cell_list=None, inplace=True):
         """
@@ -90,9 +117,11 @@ class StPipeline(object):
         :param inplace: whether inplace the original data or return a new data.
         :return:
         """
-        return filter_cells(self.data, min_gene, max_gene, min_n_genes_by_counts, max_n_genes_by_counts, pct_counts_mt,
+        data = filter_cells(self.data, min_gene, max_gene, min_n_genes_by_counts, max_n_genes_by_counts, pct_counts_mt,
                             cell_list, inplace)
+        return data
 
+    @logit
     def filter_genes(self, min_cell=None, max_cell=None, gene_list=None, inplace=True):
         """
         filter genes based on the numbers of cells.
@@ -103,8 +132,10 @@ class StPipeline(object):
         :param inplace: whether inplace the original data or return a new data.
         :return:
         """
-        return filter_genes(self.data, min_cell, max_cell, gene_list, inplace)
+        data = filter_genes(self.data, min_cell, max_cell, gene_list, inplace)
+        return data
 
+    @logit
     def filter_coordinates(self, min_x=None, max_x=None, min_y=None, max_y=None, inplace=True):
         """
         filter cells based on the coordinates of cells.
@@ -116,8 +147,10 @@ class StPipeline(object):
         :param inplace: whether inplace the original data or return a new data.
         :return:
         """
-        return filter_coordinates(self.data, min_x, max_x, min_y, max_y, inplace)
+        data = filter_coordinates(self.data, min_x, max_x, min_y, max_y, inplace)
+        return data
 
+    @logit
     def log1p(self, inplace=True, res_key='log1p'):
         """
         log1p for express matrix.
@@ -131,6 +164,7 @@ class StPipeline(object):
         else:
             self.result[res_key] = np.log1p(self.data.exp_matrix)
 
+    @logit
     def normalize_total(self, target_sum=10000, inplace=True, res_key='normalize_total'):
         """
         total count normalize the data to `target_sum` reads per cell, so that counts become comparable among cells.
@@ -145,6 +179,7 @@ class StPipeline(object):
         else:
             self.result[res_key] = normalize_total(self.data.exp_matrix, target_sum=target_sum)
 
+    @logit
     def quantile(self, inplace=True, res_key='quantile'):
         """
         Normalize the columns of X to each have the same distribution. Given an expression matrix  of M genes by N
@@ -161,6 +196,7 @@ class StPipeline(object):
         else:
             self.result[res_key] = quantile_norm(self.data.exp_matrix)
 
+    @logit
     def disksmooth_zscore(self, r=20, inplace=True, res_key='disksmooth_zscore'):
         """
         for each position, given a radius, calculate the z-score within this circle as final normalized value.
@@ -177,6 +213,7 @@ class StPipeline(object):
         else:
             self.result[res_key] = zscore_disksmooth(self.data.exp_matrix, self.data.position, r)
 
+    @logit
     def sctransform(self,
                     method="theta_ml",
                     n_cells=5000,
@@ -204,13 +241,17 @@ class StPipeline(object):
         """
         from ..preprocess.sc_transform import sc_transform
         if inplace:
-            sc_transform(self.data, method, n_cells, n_genes, filter_hvgs, res_clip_range, var_features_n)
+            self.result[res_key] = sc_transform(self.data, method, n_cells, n_genes, filter_hvgs,
+                                                res_clip_range, var_features_n)
         else:
             import copy
             data = copy.deepcopy(self.data)
             self.result[res_key] = sc_transform(data, method, n_cells, n_genes, filter_hvgs,
                                                 res_clip_range, var_features_n)
+        key = 'sct'
+        self.reset_key_record(key, res_key)
 
+    @logit
     def highly_variable_genes(self,
                          groups=None,
                          method: Optional[str] = 'seurat',
@@ -256,6 +297,8 @@ class StPipeline(object):
                                   max_disp=max_disp, min_mean=min_mean, max_mean=max_mean, span=span, n_bins=n_bins)
         hvg.fit()
         self.result[res_key] = hvg.result
+        key = 'hvg'
+        self.reset_key_record(key, res_key)
 
     def subset_by_hvg(self, hvg_res_key, inplace=True):
         """
@@ -274,6 +317,7 @@ class StPipeline(object):
         data.sub_by_index(gene_index=genes_index)
         return data
 
+    @logit
     def pca(self, use_highly_genes, n_pcs, hvg_res_key='highly_variable_genes', res_key='pca'):
         """
         Principal component analysis.
@@ -290,7 +334,8 @@ class StPipeline(object):
         x = data.exp_matrix.toarray() if issparse(data.exp_matrix) else data.exp_matrix
         res = pca(x, n_pcs)
         self.result[res_key] = pd.DataFrame(res['x_pca'])
-
+        key = 'pca'
+        self.reset_key_record(key, res_key)
     # def umap(self, pca_res_key, n_pcs=None, n_neighbors=5, min_dist=0.3, res_key='dim_reduce'):
     #     if pca_res_key not in self.result:
     #         raise Exception(f'{pca_res_key} is not in the result, please check and run the pca func.')
@@ -298,6 +343,7 @@ class StPipeline(object):
     #     res = u_map(x, 2, n_neighbors, min_dist)
     #     self.result[res_key] = pd.DataFrame(res)
 
+    @logit
     def umap(self,
              pca_res_key,
              neighbors_res_key,
@@ -350,7 +396,10 @@ class StPipeline(object):
                       min_dist=min_dist, spread=spread, n_components=n_components, maxiter=maxiter, alpha=alpha,
                       gamma=gamma, negative_sample_rate=negative_sample_rate, init_pos=init_pos)
         self.result[res_key] = pd.DataFrame(x_umap)
+        key = 'umap'
+        self.reset_key_record(key, res_key)
 
+    @logit
     def neighbors(self, pca_res_key, method='umap', metric='euclidean', n_pcs=None, n_neighbors=10, knn=True,
                   res_key='neighbors'):
         """
@@ -396,8 +445,10 @@ class StPipeline(object):
                                                          n_neighbors=n_neighbors, metric=metric, knn=knn)
         res = {'neighbor': neighbor, 'connectivities': connectivities, 'nn_dist': dists}
         self.result[res_key] = res
+        key = 'neighbors'
+        self.reset_key_record(key, res_key)
 
-    def get_neighbors_res(self, neighbors_res_key):
+    def get_neighbors_res(self, neighbors_res_key,):
         """
         get the neighbor result by the key.
 
@@ -412,6 +463,30 @@ class StPipeline(object):
         nn_dist = neighbors_res['nn_dist']
         return neighbor, connectivities, nn_dist
 
+    @logit
+    def spatial_neighbors(self, neighbors_res_key, n_neighbors=6, res_key='spatial_neighbors'):
+        """
+        Create a graph from spatial coordinates using squidpy.
+
+        :param neighbors_res_key: the key of neighbors to getting the result.
+        :param n_neighbors: Use this number of nearest neighbors.
+        :param res_key: the key for getting the result from the self.result.
+        :return:
+        """
+        from ..io.reader import stereo_to_anndata
+        import squidpy as sq
+        neighbor, connectivities, dists = copy.deepcopy(self.get_neighbors_res(neighbors_res_key))
+        adata = stereo_to_anndata(self.data)
+        sq.gr.spatial_neighbors(adata, n_neighs=n_neighbors)
+        connectivities.data[connectivities.data > 0] = 1
+        adj = connectivities + adata.obsp['spatial_connectivities']
+        adj.data[adj.data > 0] = 1
+        res = {'neighbor': neighbor, 'connectivities': adj, 'nn_dist': dists}
+        self.result[res_key] = res
+        key = 'neighbors'
+        self.reset_key_record(key, res_key)
+
+    @logit
     def leiden(self,
                neighbors_res_key,
                res_key='cluster',
@@ -444,7 +519,10 @@ class StPipeline(object):
                       use_weights=use_weights, random_state=random_state, n_iterations=n_iterations)
         df = pd.DataFrame({'bins': self.data.cell_names, 'group': clusters})
         self.result[res_key] = df
+        key = 'cluster'
+        self.reset_key_record(key, res_key)
 
+    @logit
     def louvain(self,
                 neighbors_res_key,
                 res_key='cluster',
@@ -476,7 +554,10 @@ class StPipeline(object):
                       adjacency=connectivities, flavor=flavor, directed=directed, use_weights=use_weights)
         df = pd.DataFrame({'bins': self.data.cell_names, 'group': clusters})
         self.result[res_key] = df
+        key = 'cluster'
+        self.reset_key_record(key, res_key)
 
+    @logit
     def phenograph(self, phenograph_k, pca_res_key, res_key='cluster'):
         """
         phenograph of cluster.
@@ -488,21 +569,25 @@ class StPipeline(object):
         """
         if pca_res_key not in self.result:
             raise Exception(f'{pca_res_key} is not in the result, please check and run the pca func.')
-        communities, _, _ = phe.cluster(self.result[pca_res_key], k=phenograph_k)
+        communities, _, _ = phe.cluster(self.result[pca_res_key], k=phenograph_k, clustering_algo='leiden')
         clusters = communities.astype(str)
         df = pd.DataFrame({'bins': self.data.cell_names, 'group': clusters})
         self.result[res_key] = df
+        key = 'cluster'
+        self.reset_key_record(key,res_key)
 
+    @logit
     def find_marker_genes(self,
                           cluster_res_key,
-                          method: str = 't-test',
+                          method: str = 't_test',
                           case_groups: Union[str, np.ndarray] = 'all',
                           control_groups: Union[str, np.ndarray] = 'rest',
                           corr_method: str = 'bonferroni',
                           use_raw: bool = True,
                           use_highly_genes: bool = True,
                           hvg_res_key: Optional[str] = None,
-                          res_key: str = 'marker_genes'
+                          res_key: str = 'marker_genes',
+                          output: Optional[str] = None,
                           ):
         """
         a tool of finding maker gene. for each group, find statistical test different genes between one group and
@@ -517,6 +602,7 @@ class StPipeline(object):
         :param use_highly_genes: Whether to use only the expression of hypervariable genes as input, default True.
         :param hvg_res_key: the key of highly varialbe genes to getting the result.
         :param res_key: the key for getting the result from the self.result.
+        :param output: file name for output. If None, do not generate the output file.
         :return:
         """
         from ..tools.find_markers import FindMarker
@@ -532,7 +618,18 @@ class StPipeline(object):
         tool = FindMarker(data=data, groups=self.result[cluster_res_key], method=method, case_groups=case_groups,
                           control_groups=control_groups, corr_method=corr_method)
         self.result[res_key] = tool.result
+        if output is not None:
+            import natsort
+            result = self.result[res_key]
+            show_cols = ['scores', 'pvalues', 'pvalues_adj', 'log2fc', 'genes']
+            groups = natsort.natsorted(result.keys())
+            dat = pd.DataFrame(
+                {group.split(".")[0] + "_" + key: result[group][key] for group in groups for key in show_cols})
+            dat.to_csv(output)
+        key = 'marker_genes'
+        self.reset_key_record(key, res_key)
 
+    @logit
     def spatial_lag(self,
                     cluster_res_key,
                     genes=None,
@@ -559,6 +656,7 @@ class StPipeline(object):
         tool.fit()
         self.result[res_key] = tool.result
 
+    @logit
     def spatial_pattern_score(self, use_raw=True, res_key='spatial_pattern'):
         """
         calculate the spatial pattern score.
@@ -576,3 +674,74 @@ class StPipeline(object):
         df = pd.DataFrame(x, columns=data.gene_names, index=data.cell_names)
         res = spatial_pattern_score(df)
         self.result[res_key] = res
+
+    @logit
+    def spatial_hotspot(self, use_highly_genes=True, hvg_res_key:Optional[str] = None, model='normal', n_neighbors=30,
+                        n_jobs=20, fdr_threshold=0.05, min_gene_threshold=10, outdir=None, res_key='spatial_hotspot',
+                        use_raw=True, ):
+        """
+        identifying informative genes (and gene modules)
+
+        :param use_highly_genes: Whether to use only the expression of hypervariable genes as input, default True.
+        :param hvg_res_key: the key of highly varialbe genes to getting the result.
+        :param model: Specifies the null model to use for gene expression.
+            Valid choices are:
+                - 'danb': Depth-Adjusted Negative Binomial
+                - 'bernoulli': Models probability of detection
+                - 'normal': Depth-Adjusted Normal
+                - 'none': Assumes data has been pre-standardized
+        :param n_neighbors: Neighborhood size.
+        :param n_jobs: Number of parallel jobs to run.
+        :param fdr_threshold: Correlation threshold at which to stop assigning genes to modules
+        :param min_gene_threshold: Controls how small modules can be.
+            Increase if there are too many modules being formed.
+            Decrease if substructre is not being captured
+        :param outdir: directory containing output file(hotspot.pkl). Hotspot object will be totally output here.
+            If None, results will not be output to a file.
+        :param res_key: the key for getting the result from the self.result.
+        :param use_raw: whether use the raw count express matrix for the analysis, default True.
+
+        :return:
+        """
+        from ..algorithm.spatial_hotspot import spatial_hotspot
+        if use_highly_genes and hvg_res_key not in self.result:
+            raise Exception(f'{hvg_res_key} is not in the result, please check and run the highly_var_genes func.')
+        #data = self.subset_by_hvg(hvg_res_key, inplace=False) if use_highly_genes else self.data
+        if use_raw and not self.raw:
+            raise Exception(f'self.raw must be set if use_raw is True.')
+        data = copy.deepcopy(self.raw) if use_raw else copy.deepcopy(self.data)
+        if use_highly_genes:
+            df = self.result[hvg_res_key]
+            genes_index = df['highly_variable'].values
+            gene_name = np.array(df.index)[genes_index]
+            data = data.sub_by_name(gene_name=gene_name)
+        hs = spatial_hotspot(data, model=model, n_neighbors=n_neighbors, n_jobs=n_jobs, fdr_threshold=fdr_threshold,
+                             min_gene_threshold=min_gene_threshold, outdir=outdir)
+        # res = {"results":hs.results, "local_cor_z": hs.local_correlation_z, "modules": hs.modules,
+        #        "module_scores": hs.module_scores}
+        self.result[res_key] = hs
+
+    # def scenic(self, tfs, motif, database_dir, res_key='scenic', use_raw=True, outdir=None,):
+    #     """
+    #
+    #     :param tfs: tfs file in txt format
+    #     :param motif: motif file in tbl format
+    #     :param database_dir: directory containing reference database(*.feather files) from cisTarget.
+    #     :param res_key: the key for getting the result from the self.result.
+    #     :param use_raw: whether use the raw count express matrix for the analysis, default True.
+    #     :param outdir: directory containing output files(including modules.pkl, regulons.csv, adjacencies.tsv,
+    #         motifs.csv). If None, results will not be output to files.
+    #
+    #     :return:
+    #     """
+    #     from ..algorithm.scenic import scenic as cal_sce
+    #     if use_raw and not self.raw:
+    #         raise Exception(f'self.raw must be set if use_raw is True.')
+    #     data = self.raw if use_raw else self.data
+    #     modules, regulons, adjacencies, motifs, auc_mtx, regulons_df = cal_sce(data, tfs, motif, database_dir, outdir)
+    #     res = {"modules": modules, "regulons": regulons, "adjacencies": adjacencies, "motifs": motifs,
+    #            "auc_mtx":auc_mtx, "regulons_df": regulons_df}
+    #     self.result[res_key] = res
+
+
+
